import cherrypy
import os

from py_hydropi.lib.memdatabase import MemDatabase

from py_hydropi.lib.logger import Logger

# noinspection PyProtectedMember
from cherrypy import _cpwsgi_server, _cpserver


class ApiServer(object):
    def __init__(self, db, config, service_name='cherrypy'):
        self.logger = Logger(service_name)
        self.is_running = False
        self.strict_port_checking = False
        self.adapter = None  # type: _cpserver.ServerAdapter
        self.logger = Logger('api')
        self.cherrypy_server = _cpwsgi_server.CPWSGIServer()
        if not hasattr(self, 'config'):
            self.load_config(config)

        self.db = db  # type: MemDatabase
        self.exit = None

    # noinspection PyProtectedMember
    def start(self):
        if not self.config.start:
            return
        if self.is_running:
            return
        self.is_running = True
        # hackery to stop cherrypy outputting log to stdout (spams journald when ran
        # in the foreground as a systemd service. all logs from cherrypy will now be
        # routed to syslog instead (journal if running systemd)
        cherrypy.log.screen = False
        cherrypy.log.error_log.log = self.logger.log
        cherrypy.log.error_log.propagate = False
        cherrypy._cpchecker.Checker.on = False  # stops cherrypy checking for config
        cherrypy.engine.signals.subscribe()
        cherrypy.config.update(
            {
                'global': {
                    'engine.autoreload.on': False
                }
            })
        cherrypy.log.access_log.log = self.logger.log
        cherrypy.log.access_log.propagate = False
        cherrypy.engine.start()

    def stop(self):
        self.logger.info('exiting')
        if self.is_running:
            cherrypy.engine.exit()
            self.exit = True
        self.db.server_queue.put('exit')
        self.is_running = False

    def load_config(self, config):
        self.config = config
        if not self.config.start:
            return
        self.strict_port_checking = config.strict_port_checking
        self.cherrypy_server.bind_addr = (config.listen_address,
                                          config.port)
        self.lazy_load_endpoints()
        self.adapter = _cpserver.ServerAdapter(cherrypy.engine,
                                               self.cherrypy_server,
                                               self.cherrypy_server.bind_addr)
        self.adapter.subscribe()

    def lazy_load_endpoints(self, endpoint_dir=None):
        if endpoint_dir is None:
            endpoint_dir = (os.path.join(os.path.dirname(os.path.realpath(__file__)), 'endpoints'))
        for found in os.listdir(endpoint_dir):
            _endpoint_dir = os.path.join(endpoint_dir, found)
            if found not in ['__init__.py', '__pycache__']:
                if os.path.isdir(_endpoint_dir):
                    self.lazy_load_endpoints(_endpoint_dir)
                else:
                    endpoint = '/' + str(_endpoint_dir.split('endpoints/')[1]).replace(found, '')
                    imp_dir = 'py_hydropi' + str(
                        _endpoint_dir).split('py_hydropi')[2].replace('/', '.').replace('.py', '')
                    mod = __import__(str(imp_dir), fromlist=[found])
                    found = ''.join([str(i.capitalize()) for i in found.replace('.py', '').split('_')])
                    class_ = getattr(mod, found)
                    instance = class_(self)
                    cherrypy.tree.mount(instance, endpoint,
                                        {
                                            '/': {
                                                'request.dispatch': cherrypy.dispatch.MethodDispatcher(),
                                                'tools.response_headers.on': True,
                                                'tools.response_headers.headers': [('Content-Type', 'text/plain')]
                                            }
                                        })



import json
import subprocess

from py_hydropi.lib.modules.inputs import Input


class HomeLabPH(Input):
    provides = ('HOMELAB_PH', 'HOMELABPH')
    get_ph_command = '/var/www/homelab/cgi-bin/get_pH.sh -j'
    frequency = 30

    def __init__(self, value_index=None):
        super().__init__()
        self.value_index = value_index

    def _read(self):
        try:
            e = subprocess.check_output(self.get_ph_command, shell=True)
            data = json.loads(e.decode('utf-8'))
            self._value = data.get(self.value_index)

        except Exception as e:
            self.logger.error(e)



import os
import time

if os.environ.get('PY_HYDROPI_TESTING', '').lower() == 'true':
    class Adafruit_DHT:
        value = 10, 10

        @classmethod
        def read_retry(cls, _, channel):
            return cls.value
else:
    import Adafruit_DHT

from py_hydropi.lib.modules.inputs import Input


class DHTxxInput(Input):
    sensor_type = None

    def __init__(self, channel=None, value_index=None, power_channel=None, pi_timer=None, **kwargs):
        super().__init__(**kwargs)
        if power_channel is not None and value_index == 0:
            assert pi_timer is not None
            self.gpio = pi_timer.gpio
            self.power_channel = int(power_channel)
            self.gpio.setup_output_channel(self.power_channel)
            self.gpio.set_output_on(self.power_channel)

        assert channel and value_index is not None
        self.channel = channel
        self.value_index = value_index

    def _read(self):
        val = None
        try:
            val = Adafruit_DHT.read_retry(self.sensor_type, self.channel)
            val = val[self.value_index]
        except IndexError:
            self.logger.error('error processing response from sensor channel {}({})'.format(
                self.channel, self.value_index))
        except Exception as e:
            self.logger.error('{} error reading from {}: channel {}({})'.format(
                e, self.__class__.__name__, self.channel, self.value_index))
        if val is None:
            self.logger.error('{}: {}({}) returned None'.format(
                self.__class__.__name__, self.channel, self.value_index))
        self.power_cycle()
        return val

    def power_cycle(self):
        if hasattr(self, 'power_channel'):
            self.gpio.set_output_off(self.power_channel)
            time.sleep(1)
            self.gpio.set_output_on(self.power_channel)


class DHT11Input(DHTxxInput):
    provides = ('DHT11', 'DHT-11')
    sensor_type = 11
    frequency = 10


class DHT22Input(DHTxxInput):
    provides = ('DHT22', 'DHT-22')
    sensor_type = 22
    frequency = 10



import os
import time

import requests

from py_hydropi.lib.logger import Logger


SUCCESS = 204


class _InfluxDBClient:
    def __init__(self, endpoint="127.0.0.1", port=8086, hostname=None, db='py_hydropi', timeout=0.5):
        self.logger = Logger(self.__class__.__name__)
        self.endpoint = endpoint
        self.port = port
        self.hostname = hostname or self._get_hostname()
        self.db = db
        self.timeout = timeout

    @staticmethod
    def _get_hostname():
        with open('/etc/hostname') as f:
            return f.read()

    @staticmethod
    def _get_current_timestamp():
        return str(time.time()).replace('.', '')

    def push(self, callback, metric, value):
        self.logger.debug('metric: {}, value: {}'.format(metric, value))
        if value is not None:
            payload = "{metric},host={hostname} value={value} {timestamp}00".format(
                metric=metric,
                hostname=self.hostname,
                value=value,
                timestamp=self._get_current_timestamp()
            )
            result = requests.post(url='http://{}:{}/write?db={}'.format(self.endpoint, self.port, self.db),
                                   data=payload,
                                   headers={'Content-Type': 'application/octet-stream'}, timeout=self.timeout)
            if result.status_code != SUCCESS:
                callback([result.status_code, result.reason])


if os.environ.get('PY_HYDROPI_TESTING', '').lower() == 'true':
    class InfluxDBClient(_InfluxDBClient):
        def push(self, callback, metric, value):
            self.logger.debug('metric: {}, value: {}'.format(metric, value))
            if value is not None:
                callback([200, 'success'])

else:
    InfluxDBClient = _InfluxDBClient



import os

from py_hydropi.lib.logger import Logger


class Output:
    types = ('lights', 'water_pumps', 'air_pumps')

    def __init__(self, pi_timer, channel: int, initial_state=False, output_type=None):
        self.logger = Logger(self.__class__.__name__)
        self.pi_timer = pi_timer
        self.gpio = self.pi_timer.gpio
        self.channel = channel
        self.type = output_type
        self.state = initial_state
        self.gpio.setup_output_channel(self.channel, initial_state=initial_state)
        self.manual_control = False

    def activate(self):
        """
        power the GPIO channel associated with this output
        """
        if not self.state:
            self.logger.debug('activate signal recieved for channel {}. Signalling GPIO'.format(self.channel))
            self.state = True
            self.gpio.set_output_on(self.channel)

    def deactivate(self):
        """
        de-power the GPIO channel associated with this output
        """
        if self.state:
            self.logger.debug('deactivate signal recieved for channel {}. Signalling GPIO'.format(self.channel))
            self.state = False
            self.gpio.set_output_off(self.channel)

    def set_state(self, state):
        if state:
            self.activate()
        else:
            self.deactivate()

    def to_json(self):
        return {'channel': self.channel, 'state': self.state}



from threading import Thread

from py_hydropi.lib.logger import Logger


class ThreadedDaemon:
    def __init__(self):
        self.logger = Logger(self.__class__.__name__)
        self._thread = None  # type: Thread
        self._continue = False  # set to false to exit self._timer_loop

    def start(self):
        self._continue = True
        self._thread = Thread(target=self._main_loop)
        self._thread.start()
        return self

    def stop(self):
        self.logger.info('recieved exit command, stopping terminating daemon main loop')
        self._continue = False
        if self._thread.is_alive():
            self._thread.join(timeout=self.frequency if hasattr(self, 'frequency') else 20)

    def _main_loop(self):
        raise NotImplementedError

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



import os

from .logger import Logger

if os.environ.get('PY_HYDROPI_TESTING', '').lower() == 'true':
    # noinspection PyUnusedLocal
    class RPIGPIO:
        class Bcm:
            pass

        class In:
            pass

        class Out:
            pass

        class High:
            pass

        class Low:
            pass

        BCM = Bcm()
        IN = In()
        OUT = Out()
        HIGH = High()  # OFF
        LOW = Low()    # ON

        @staticmethod
        def setmode(mode):
            assert isinstance(mode, RPIGPIO.Bcm)

        @staticmethod
        def cleanup():
            return

        @staticmethod
        def setup(channel, input_output, pull_up_down=None, initial=None):
            assert type(channel) == int
            assert isinstance(input_output, (RPIGPIO.In, RPIGPIO.Out, RPIGPIO.High, RPIGPIO.Low))

        @staticmethod
        def output(channel, state):
            assert type(channel) == int
            assert isinstance(state, (RPIGPIO.High, RPIGPIO.Low))

        @staticmethod
        def input(channel):
            assert type(channel) == int
else:
    import RPi.GPIO as RPIGPIO


# noinspection PyUnusedLocal
class GPIO(object):
    def __init__(self):
        self._GPIO = RPIGPIO  # type: RPIGPIO
        self._GPIO.setmode(self._GPIO.BCM)
        self.logger = Logger('GPIO')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._GPIO.cleanup()

    def setup_input_channel(self, channel: int, pull_up_down=None):
        self._GPIO.setup(int(channel), self._GPIO.IN)

    def setup_output_channel(self, channel: int, initial_state=False):
        try:
            self._GPIO.setup(int(channel), self._GPIO.OUT)
            self.logger.debug('set channel {} as output'.format(channel))
        except ValueError:
            self.logger.error('error processing channel {}'.format(channel))

    def set_output_on(self, channel):
        try:
            self._GPIO.output(channel, self._GPIO.LOW)
        except RuntimeError:
            self.logger.error('error setting channel {} on'.format(channel))

    def set_output_off(self, channel):
        try:
            self._GPIO.output(channel, self._GPIO.HIGH)
        except RuntimeError:
            self.logger.error('error setting channel {} off'.format(channel))

    def get_input(self, channel):
        return self._GPIO.input(channel)

    def cleanup(self):
        self._GPIO.cleanup()



class OutputMetricCollector:
    def __init__(self, db, reporter):
        super().__init__()
        self.db = db
        self.reporter = reporter

    def push_all(self):
        outputs = []
        for type_, group_dict in self.db.controllers.items():

            for group, controller in group_dict.items():
                if hasattr(controller, '_rising_object') and hasattr(controller, '_falling_object'):
                    outputs.extend([controller._rising_object, controller._falling_object])
                else:
                    if hasattr(controller, 'attached_triggered_outputs'):
                        for d in controller.attached_triggered_outputs.values():
                            outputs.extend(d.get('objects'))
                    outputs.extend(controller.attached_outputs)

        return [('channel{channel}activated'.format(channel=output.channel), int(output.state))
                for output in outputs]



import yaml
import os


MODULE_CONFIG_FILENAME = 'module_config.yaml'
API_CONFIG_FILENAME = 'api_config.yaml'
METRICS_CONFIG_FILENAME = 'metrics.yaml'

if os.environ.get('PY_HYDROPI_TESTING', '').lower() == 'true':
    default_config_dir = 'py_hydropi/defaults/'
else:
    default_config_dir = '/etc/py_hydropi/'


class BaseConfig(object):
    def __init__(self, config_dir=default_config_dir):
        assert '.yaml' in config_dir

        path = '/'.join(a for a in config_dir.split('/') if not a.endswith('.yaml'))
        self.filename = '/'.join(a for a in config_dir.split('/') if a.endswith('.yaml'))
        self.config_dir = path

        if os.path.isdir(self.config_dir):
            self._load_config()
        else:
            os.mkdir(self.config_dir)
            # TODO: generate default config file

    def _load_config(self):
        with open(self.config_dir + '/' + self.filename, 'r') as config_yaml:
            self.config = yaml.load(config_yaml)
            self._config_parser()

    def _config_parser(self):
        pass


class ModuleConfig(BaseConfig):
    def __init__(self, config_dir=default_config_dir):
        config_dir += MODULE_CONFIG_FILENAME
        super().__init__(config_dir)


class MetricsConfig(BaseConfig):
    def __init__(self, config_dir=default_config_dir):
        config_dir += METRICS_CONFIG_FILENAME
        self.enabled = False
        super().__init__(config_dir)

    def _config_parser(self):
        if self.config.get('reporter'):
            self.enabled = True
            self.reporter = list(self.config.get('reporter').keys())[0]


class ApiConfig(BaseConfig):
    def __init__(self, config_dir=default_config_dir):
        config_dir += API_CONFIG_FILENAME
        super().__init__(config_dir)

    def _config_parser(self):
        self.strict_port_checking = self.config.get('strict_port_checking')
        self.listen_address = self.config.get('listen_address')
        self.port = self.config.get('port')
        self.start = self.config.get('start')



from datetime import datetime, timedelta, time
from time import sleep

from py_hydropi.lib import Output
from py_hydropi.lib.modules.switch import Switch
from ..time_utils import parse_clock_time_string, parse_simple_time_string, time_to_datetime


class Timer(Switch):
    def __init__(self):
        super().__init__()
        self.attached_triggered_outputs = {}
        self.activated_time = None  # type: datetime
        self.deactivated_time = None  # type: datetime
        self.triggered_outputs_activated = False

    @property
    def all_outputs(self):
        out = super().all_outputs
        for outputs in self.attached_triggered_outputs.values():
            out.extend(outputs.get('objects'))
        return out

    def _activate_objects(self, activated_time=None):
        if not self.outputs_activated:
            self.activated_time = activated_time or datetime.now()
            self.deactivated_time = None
            super()._activate_objects()

    def _deactivate_objects(self, deactivaed_time=None):
        if self.outputs_activated:
            self.activated_time = None
            self.deactivated_time = deactivaed_time or datetime.now()
            super()._deactivate_objects()

    def _activate_triggered_objects(self, group_name):
        if not self.triggered_outputs_activated:
            for output in self.attached_triggered_outputs[group_name]['objects']:
                if output.manual_control:
                    self.logger.info('output {} is manually controlled, skipping signal'.format(output.channel))
                else:
                    self.logger.info('signalling triggered output {} to activate'.format(output.channel))
                    output.activate()
                self.triggered_outputs_activated = True

    def _deactivate_triggered_objects(self, group_name):
        if self.triggered_outputs_activated:
            for output in self.attached_triggered_outputs[group_name]['objects']:
                if output.manual_control:
                    self.logger.info('output {} is manually controlled, skipping signal'.format(output.channel))
                else:
                    self.logger.info('signalling triggered output {} to deactivate'.format(output.channel))
                    output.deactivate()
                self.triggered_outputs_activated = False

    def attach_triggered_object(self, obj, group_name, before, after):
        if group_name not in self.attached_triggered_outputs.keys():
            self.attached_triggered_outputs[group_name] = {
                'objects': [],
                'before': parse_simple_time_string(before),
                'after': parse_simple_time_string(after)
            }

        if type(obj) in (list, tuple):
            self.attached_triggered_outputs[group_name]['objects'].extend(obj)
        else:
            self.attached_triggered_outputs[group_name]['objects'].append(obj)

        return self

    def _check_timer(self):
        """
        extend this function in inherited timer classes to check
        timer conditions and sets output status accordingly
        :return: nothing
        """
        raise NotImplementedError()

    def _check_before_trigger(self):
        raise NotImplementedError()

    def _check_after_trigger(self):
        raise NotImplementedError()

    def _main_loop(self):
        while self._continue:
            self._check_timer()
            self._check_after_trigger()
            self._check_before_trigger()
            sleep(0.1)

    def stop(self):
        for group in self.attached_triggered_outputs.keys():
            self._deactivate_triggered_objects(group)
        super(Timer, self).stop()

    def to_json(self):
        return {
            'attached_outputs': [o.to_json() for o in self.attached_outputs],
            'attached_triggered_outputs': self._decode_attached_triggered_outputs(),
            'active': self._continue,
            'outputs_activated': self.outputs_activated,
            'triggered_outputs_activated': self.triggered_outputs_activated
        }

    def _decode_attached_triggered_outputs(self):
        out_dict = {}
        for group_name, param_dict in self.attached_triggered_outputs.items():
            out_dict[group_name] = {'before': param_dict.get('before'),
                                    'after': param_dict.get('after'),
                                    'objects': [o.to_json() for o in param_dict.get('objects')]
                                    }
        return out_dict

    @classmethod
    def load_config(cls, raspberry_pi_timer, config):
        raise NotImplementedError


class ClockTimer(Timer):
    def __init__(self, active_hours):
        self.on_time, self.off_time = parse_clock_time_string(active_hours)
        super().__init__()

    @staticmethod
    def _get_current_time():
        return datetime.now()

    @staticmethod
    def time_in_range(start, end, x):
        """Return true if x is in the range [start, end]"""
        if start <= end:
            return start <= x <= end
        else:
            return start <= x or x <= end

    def _check_timer(self):
        now = self._get_current_time().time()
        on = time_to_datetime(now, self.on_time).time()
        off = time_to_datetime(now, self.off_time).time()
        if self.outputs_activated:
            if self.time_in_range(off, on, now):
                self._deactivate_objects(now)
        else:
            if self.time_in_range(on, off, now):
                self._activate_objects(now)

    def _check_before_trigger(self):
        for group_name, opts in self.attached_triggered_outputs.items():
            now = datetime.now()
            on_string = now.strftime('%b %d %Y {}'.format(self.on_time))
            on_datetime = datetime.strptime(on_string, '%b %d %Y %I:%M%p')
            if on_datetime < now:
                on_datetime += timedelta(days=1)
            if on_datetime - timedelta(seconds=opts['before']) > now:
                self._activate_triggered_objects(group_name)

    def _check_after_trigger(self):
        for group_name, opts in self.attached_triggered_outputs.items():
            now = datetime.now()
            off_string = now.strftime('%b %d %Y {}'.format(self.off_time))
            off_datetime = datetime.strptime(off_string, '%b %d %Y %I:%M%p')
            if off_datetime < now:
                off_datetime += timedelta(days=1)
            if off_datetime + timedelta(seconds=opts['after']) > now:
                self._deactivate_triggered_objects(group_name)

    @classmethod
    def load_config(cls, raspberry_pi_timer, config):
        return {group: cls(
            active_hours=group_settings.get('active_hours')
        ).attach_object(
            [Output(raspberry_pi_timer, chan)
             for chan in group_settings.get('channels')]
        ) for group, group_settings in config.items()}


class SimpleTimer(Timer):
    def __init__(self, on_time, off_time):
        self.on_time = parse_simple_time_string(on_time)  # type: int
        self.off_time = parse_simple_time_string(off_time)  # type: int
        super().__init__()

    @classmethod
    def load_config(cls, raspberry_pi_timer, config):
        return {group: cls(
            on_time=group_settings.get('on_time'),
            off_time=group_settings.get('off_time')
        ).attach_object(
            [Output(raspberry_pi_timer, chan)
             for chan in group_settings.get('channels')]
        ) for group, group_settings in config.items()}

    def _check_timer(self):
        now = datetime.now()
        if not self.outputs_activated:
            if self.deactivated_time is None or self.deactivated_time + timedelta(seconds=self.off_time) < now:
                self.activated_time = datetime.now()
                self._activate_objects()

        else:
            if self.activated_time is None or self.activated_time + timedelta(seconds=self.on_time) < now:
                self.deactivated_time = datetime.now()
                self._deactivate_objects()

    def _check_before_trigger(self):
        for group_name, opts in self.attached_triggered_outputs.items():
            if not self.triggered_outputs_activated:
                if self.deactivated_time:
                    now = datetime.now()
                    if self.deactivated_time + timedelta(seconds=self.off_time) - timedelta(seconds=opts['before']) < now:
                        self._activate_triggered_objects(group_name)

    def _check_after_trigger(self):
        for group_name, opts in self.attached_triggered_outputs.items():
            if self.triggered_outputs_activated:
                if self.activated_time:
                    now = datetime.now()
                    if self.activated_time + timedelta(seconds=self.on_time) + timedelta(seconds=opts['after']) < now:
                        self._deactivate_triggered_objects(group_name)



from py_hydropi.lib.API.main import ApiServer
import cherrypy


# noinspection PyPep8Naming
class APIEndpointBase(object):
    exposed = True
    error_405 = cherrypy.HTTPError(405, None)

    def __init__(self, api):
        self.api = api  # type: ApiServer

    @cherrypy.tools.json_out()
    def GET(self):
        return self._get()

    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out()
    def POST(self):
        return self._post(cherrypy.request.json)

    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out()
    def PUT(self):
        return self._put(cherrypy.request.json)

    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out()
    def DELETE(self):
        return self._delete(cherrypy.request.json)

    def _delete(self, *args, **kwargs):
        raise self.error_405

    def _put(self, *args, **kwargs):
        raise self.error_405

    def _post(self, *args, **kwargs):
        raise self.error_405

    def _get(self):
        raise self.error_405



from py_hydropi.lib.modules.inputs import Input


class OneWireInput(Input):
    provides = ('DS18B20', )
    _parsers = {'28-041752029bff': lambda i: int(i.splitlines()[1].split('t=')[1]) / 1000.0}
    _path = '/sys/bus/w1/devices'
    _sensor_template = '{path}/{sensor_id}/w1_slave'
    frequency = 10

    def __init__(self, sensor_id=None, **kwargs):
        super().__init__(**kwargs)
        assert sensor_id

        self.sensor_id = sensor_id
        self._sensor_path = self._sensor_template.format(path=self._path,
                                                         sensor_id=self.sensor_id)

    def _read(self):
        try:
            with open(self._sensor_path) as s:
                try:
                    return self._parsers[self.sensor_id](s.read())

                except KeyError:
                    self.logger.error('parser not defined for: {}'.format(self.sensor_id))

                except IndexError:
                    self.logger.error('error reading sensor data')
                    return

        except FileNotFoundError:
            self.logger.error('specified sensor not found: {}\nexiting monitoing thread'.format(self._sensor_path))
            self._continue = False



import time
from distutils.core import setup
import os

from setuptools import find_packages

required_packages = [
    'PyYAML',
    'cherrypy',
    'systemd',
    'pep8',
    'requests'
]

is_rasp_pi = os.uname()[4].startswith('arm')
if is_rasp_pi:
    required_packages.extend([
        'RPi.GPIO',
        'Adafruit_Python_DHT',
        'hcsr04sensor'
    ])

VERSION = os.environ.get('TRAVIS_TAG') or '0.0.0.post{}'.format(int(time.time()))

setup(
    name='py_hydropi',
    version='0.0.10.post1542299571',
    url='https://github.com/chestm007/py_hydropi',
    packages=find_packages(),
    author='max',
    author_email='chestm007@hotmail.com',
    license='GPL-2.0',
    requires=required_packages,
    include_package_data=True,
    description='Hydroponics controller for RaspberryPi',
    install_requires=required_packages,
    entry_points="""
        [console_scripts]
        py_hydropi=py_hydropi.main:main
    """
)



import re
from datetime import datetime

SECONDS = 's'
MINUTES = 'm'
HOURS = 'h'
DAYS = 'd'


def parse_clock_time_string(string):
    on, off = string.split(' - ')
    return on, off


def parse_simple_time_string(string):
    total_secs = 0
    time_pattern = (
        '((?P<d>\d+){0})?((?P<h>\d+){1})?((?P<m>\d+){2})?((?P<s>\d+){3})?'.format(DAYS, HOURS, MINUTES, SECONDS))
    pattern = re.compile(time_pattern)

    found = pattern.search(string)
    if found:
        if found.group(SECONDS):
            total_secs += int(found.group(SECONDS))
        if found.group(MINUTES):
            total_secs += int(found.group(MINUTES)) * 60
        if found.group(HOURS):
            total_secs += int(found.group(HOURS)) * 60 * 60
        if found.group(DAYS):
            total_secs += int(found.group(DAYS)) * 24 * 60 * 60
    return total_secs


def time_to_datetime(now, time):
    time_string = now.strftime('%b %d %Y {}'.format(time))
    return datetime.strptime(time_string, '%b %d %Y %I:%M%p')



import signal
import sys

from multiprocessing import Queue

import time

from py_hydropi.lib.config import ApiConfig, MetricsConfig
from py_hydropi.lib.iter_utils import JSONableDict
from py_hydropi.lib.memdatabase import MemDatabase
from py_hydropi.lib.metrics.collector_controller import MetricCollectorController
from py_hydropi.lib.metrics.collectors.output import OutputMetricCollector
from py_hydropi.lib.metrics.collectors.sensor import SensorMetricCollector
from py_hydropi.lib.modules.inputs import Input
from py_hydropi.lib.modules.threshold_switch import ThresholdSwitch
from py_hydropi.lib.modules.timer import SimpleTimer, ClockTimer


from py_hydropi.lib import Logger, Output, ModuleConfig, GPIO
from py_hydropi.lib.API.main import ApiServer


class RaspberryPiTimer(object):
    def __init__(self):
        self.logger = Logger(self.__class__.__name__)
        self.gpio = GPIO()
        self.queue = Queue()
        self.db = MemDatabase(self.queue)
        self.db.gpio = self.gpio

        self.module_config = ModuleConfig()
        self.api_config = ApiConfig()
        self.metrics_config = MetricsConfig()

        if self.api_config.start:
            self.api = ApiServer(self.db, self.api_config)

        self.setup_io()

        if self.metrics_config.enabled:
            self.metrics_controller = MetricCollectorController(
                self.db, self.metrics_config, [
                    SensorMetricCollector, OutputMetricCollector
                ]
            )
        signal.signal(signal.SIGINT, self.stop)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def cleanup(self):
        self.gpio.cleanup()

    def start(self):
        self.logger.info('starting all services')
        if hasattr(self, 'api'):
            self.logger.info('starting API...')
            self.api.start()

        self.logger.info('starting controllers...')
        for type_ in self.db.controllers.values():
            for name, c in type_.items():
                self.logger.info('Starting controller: {}'.format(name))
                c.start()
        if hasattr(self, 'metrics_controller'):
            self.logger.info('starting metrics reporting...')
            self.metrics_controller.start()

        self._queue_loop()

    def _queue_loop(self):
        while True:
            try:
                recv_req = self.queue.get()
                if recv_req == 'exit':
                    self.logger.info('recieved exit command. shutting down services.')
                    break
            except KeyboardInterrupt:
                self.logger.info('Detected KeyboardInterupt in queue loop, exiting gracefully')
                self.stop()
                break

    def stop(self, *_):
        self.logger.info('stopping all services')
        if hasattr(self, 'api'):
            self.logger.info('Shutting down API server')
            self.api.stop()

        for type_ in self.db.controllers.values():
            for name, c in type_.items():
                self.logger.info('Stopping controller: {}'.format(name))
                c.stop()

        if hasattr(self, 'metrics_controller'):
            self.logger.info('Shutting down metrics collector.')
            self.metrics_controller.stop()

        self.logger.info('Terminating input reader loops.')
        for sensor in self.db.inputs.values():
            sensor.stop()

        self.cleanup()
        self.logger.info('system shutdown complete')

    def setup_io(self):
        self.logger.info('loading sensors...')
        sensor_config = self.module_config.config.get('sensors')
        if sensor_config is not None:
            self.db._inputs = Input.load_config(self, sensor_config)
            for sensor in self.db.inputs.values():
                sensor.start()

        self.logger.info('loading timers...')
        clock_config = self.module_config.config.get('clock_timer')
        if clock_config is not None:
            self.db.controllers['clock_timer'] = JSONableDict(ClockTimer.load_config(self, clock_config))

        simple_config = self.module_config.config.get('simple_timer')
        if simple_config is not None:
            self.db.controllers['simple_timer'] = JSONableDict(SimpleTimer.load_config(self, simple_config))

        self.logger.info('loading switches...')
        threshold_config = self.module_config.config.get('threshold')
        if threshold_config is not None:
            self.db.controllers['threshold'] = JSONableDict(ThresholdSwitch.load_config(self, threshold_config))

        triggered_config = self.module_config.config.get('triggered')
        if triggered_config is not None:
            for group, group_settings in triggered_config.items():
                parent_controller, parent_group = group_settings.get('object').split('.')
                parent_controller = self.db.controllers.get(parent_controller)
                if parent_controller:
                    parent_group = parent_controller.get(parent_group)
                    if parent_group:
                        parent_group.attach_triggered_object(
                            obj=[Output(self, chan)
                                 for chan in group_settings.get('channels')],
                            group_name=group,
                            before=group_settings.get('before'),
                            after=group_settings.get('after'))
                        continue
                self.logger.error('error binding trigger: {}'.format(group))


def main():
    rpi_timer = RaspberryPiTimer()
    rpi_timer.start()


if __name__ == '__main__':
    main()



from py_hydropi.lib.iter_utils import JSONableDict, JSONableList


class MemDatabase(object):
    def __init__(self, queue):
        self.controllers = JSONableDict()
        self.gpio = None
        self.server_queue = queue
        self._inputs = {}

    def get_input(self, sensor_id=None):
        return self._inputs.get(sensor_id)

    def get_output(self, output_id):
        type_, name_ = output_id.split('.')
        group = self.controllers.get(type_)
        if group:
            return group.get(name_)

    def outputs(self):
        outputs = JSONableList()
        for type_, value in self.controllers.items():
            for name_, switch in value.items():
                if hasattr(switch, 'all_outputs'):
                    outputs.extend_if_not_in(switch.all_outputs)
        return outputs

    @property
    def inputs(self):
        return self._inputs



import time
import os

from py_hydropi.lib.iter_utils import avg
from py_hydropi.lib.threaded_daemon import ThreadedDaemon


class Input(ThreadedDaemon):
    frequency = 60

    def __init__(self, samples=1, value_processor=None, **kwargs):
        super().__init__()

        self._samples = samples
        self._last_value = 0
        if not hasattr(self, '_value'):
            self._value = None

        if value_processor is not None:
            if value_processor.get('range_percentage'):
                range_percentage = value_processor.get('range_percentage')
                self.rp_min = range_percentage.get('min')
                self.rp_max = range_percentage.get('max')
                self.rp_inverted = range_percentage.get('inverted')
                if self.rp_inverted:
                    self.value_processor = lambda v: 100 - ((v - self.rp_min) / (self.rp_max - self.rp_min)) * 100
                else:
                    self.value_processor = lambda v: ((v - self.rp_min) / (self.rp_max - self.rp_min)) * 100
        else:
            self.value_processor = lambda v: v

    @property
    def value(self):
        return self._value

    @staticmethod
    def load_config(pi_timer, config):

        sensors = {}
        for sensor, config in config.items():
            def create_dhtxx_sensor(variant):
                return variant(channel=config.get('channel'),
                               value_index=i,
                               power_channel=config.get('power_channel'),
                               pi_timer=pi_timer)

            type_ = config.get('type', '').replace('-', '_').upper()

            if type_ in DHT11Input.provides or type_ in DHT22Input.provides:
                if type_ in DHT11Input.provides:
                    dht_variant = DHT11Input
                else:
                    dht_variant = DHT22Input
                for i, val in enumerate(config.get('provides')):
                    sensors['{}.{}'.format(sensor, val)] = create_dhtxx_sensor(dht_variant)

            elif type_ in OneWireInput.provides:
                sensors[sensor] = OneWireInput(**config)

            elif type_ in UltrasonicInput.provides:
                sensors[sensor] = UltrasonicInput(pi_timer=pi_timer, **config)

            elif type_ in HomeLabPH.provides:
                sensors['{}.pH'.format(sensor)] = HomeLabPH(value_index='pH')
                sensors['{}.temperature'.format(sensor)] = HomeLabPH(value_index='t')
        return sensors

    def _main_loop(self):
        while self._continue:
            vals = []
            for i in range(self._samples):
                v = self._read()
                if v:
                    vals.append(self.value_processor(v))
            if vals:
                self._value = avg(vals)
            time.sleep(self.frequency)

    def _read(self):
        raise NotImplementedError

    def to_json(self):
        return 'foo'


from py_hydropi.lib.modules.sensors.dhtxx import DHT11Input, DHT22Input
from py_hydropi.lib.modules.sensors.one_wire import OneWireInput
from py_hydropi.lib.modules.sensors.hc_sr04 import UltrasonicInput
from py_hydropi.lib.modules.sensors.homelab_ph import HomeLabPH
